/* @(#)geterrno.c	1.9 03/06/15 Copyright 1985, 1995-2003 J. Schilling */
/*
 *	Get error number
 *
 *	Copyright (c) 1985, 1995-2003 J. Schilling
 */
/*
 * Copyright Jörg Schilling. All rights reserved.
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only.
 * See the file CDDL.Schily.txt in this distribution or
 * http://opensource.org/licenses/cddl1.php for details.
 */

#include <errno.h>
#include <standard.h>
#include <schily.h>

#ifndef	HAVE_ERRNO_DEF
extern	int	errno;
#endif

#ifdef	geterrno
#undef	geterrno
#endif

EXPORT int
geterrno()

{
	return (errno);
}
