/* @(#)ffilewrite.c	1.6 04/08/08 Copyright 1986 J. Schilling */
/*
 *	Copyright (c) 1986 J. Schilling
 */
/*
 * Copyright Jörg Schilling. All rights reserved.
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only.
 * See the file CDDL.Schily.txt in this distribution or
 * http://opensource.org/licenses/cddl1.php for details.
 */

#include "schilyio.h"

EXPORT int
ffilewrite(f, buf, len)
	register FILE	*f;
	void	*buf;
	int	len;
{
	down2(f, _IORWT, _IORW);

	return (write(fileno(f), (char *)buf, len));
}
