#ident "@(#)os-ms-dos.def	1.1 03/11/23 "
###########################################################################
#
# Global os definitions for MS-DOS/DJGPP v2
#
###########################################################################
# Copyright Jörg Schilling. All rights reserved.
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License, Version 1.0 only.
# See the file CDDL.Schily.txt in this distribution or
# http://opensource.org/licenses/cddl1.php for details.
###########################################################################
MANSTYLE=	sysv
