/* @(#)xtab.h	1.1 03/10/03 Copyright 2001-2003 J. Schilling */
/*
 *	Copyright (c) 2001-2003 J. Schilling
 */
/*
 * Copyright Jörg Schilling. All rights reserved.
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only.
 * See the file CDDL.Schily.txt in this distribution or
 * http://opensource.org/licenses/cddl1.php for details.
 */

#ifndef	_XTAB_H
#define	_XTAB_H

#ifndef _MCONFIG_H
#include <mconfig.h>
#endif
typedef	void (*xt_function)	__PR((FINFO *, char *, int, char *, int));

typedef struct {
	char		*x_name;
	int		x_namelen;
	xt_function	x_func;
	int		x_flag;
} xtab_t;

#endif	/* _XTAB_H */
