#ident @(#)man-bsd.def	1.1 97/02/11 
###########################################################################
# Written 1997 by J. Schilling
###########################################################################
#
# Definition for manual sections on BSD based systems
#
###########################################################################
# Copyright Jörg Schilling. All rights reserved.
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License, Version 1.0 only.
# See the file CDDL.Schily.txt in this distribution or
# http://opensource.org/licenses/cddl1.php for details.
###########################################################################

MANSECT_CMD=		man1
MANSECT_SYSCALL=	man2
MANSECT_LIB=		man3
MANSECT_FILEFORM=	man5
MANSECT_HDR=		man7
MANSECT_TABLES=		man7
MANSECT_MACROS=		man7
MANSECT_GAMES=		man6
MANSECT_DEMOS=		man6
MANSECT_DEVICE=		man4
MANSECT_NETWORK=	man4
MANSECT_ADMIN=		man8
MANSECT_DRIVER=		man4

MANSUFF_CMD=		1
MANSUFF_SYSCALL=	2
MANSUFF_LIB=		3
MANSUFF_FILEFORM=	5
MANSUFF_HDR=		7
MANSUFF_TABLES=		7
MANSUFF_MACROS=		7
MANSUFF_GAMES=		6
MANSUFF_DEMOS=		6
MANSUFF_DEVICE=		4
MANSUFF_NETWORK=	4
MANSUFF_ADMIN=		8
MANSUFF_DRIVER=		4
