/* @(#)getcwd.h	1.3 01/07/15 Copyright 1998 J. Schilling */
/*
 *	Definitions for getcwd()
 *
 *	Copyright (c) 1998 J. Schilling
 */
/*
 * Copyright Jörg Schilling. All rights reserved.
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only.
 * See the file CDDL.Schily.txt in this distribution or
 * http://opensource.org/licenses/cddl1.php for details.
 */

#ifndef	_GETCWD_H
#define	_GETCWD_H

#ifndef _MCONFIG_H
#include <mconfig.h>
#endif

#ifdef JOS
#	ifndef	_INCL_SYS_STYPES_H
#	include <sys/stypes.h>
#	define	_INCL_SYS_STYPES_H
#	endif
	extern char	*gwd();
#	define		getcwd(b, len)	gwd(b)
#else
#	ifndef	HAVE_GETCWD
#		define	getcwd(b, len)	getwd(b)
#	endif
#endif

#endif	/* _GETCWD_H */
