/* @(#)filemopen64.c	1.2 04/08/08 Copyright 2000-2004 J. Schilling */
/*
 *	Copyright (c) 2000-2004 J. Schilling
 */
/*
 * Copyright Jörg Schilling. All rights reserved.
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only.
 * See the file CDDL.Schily.txt in this distribution or
 * http://opensource.org/licenses/cddl1.php for details.
 */

#define	USE_LARGEFILES

#include "schilyio.h"

/*#if	_LFS_LARGEFILE*/
#ifdef	HAVE_LARGEFILES

#include "filemopen.c"

#endif
