/* @(#)filemopen.c	1.3 04/08/08 Copyright 1986, 1995-2004 J. Schilling */
/*
 *	Copyright (c) 1986, 1995-2004 J. Schilling
 */
/*
 * Copyright Jörg Schilling. All rights reserved.
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only.
 * See the file CDDL.Schily.txt in this distribution or
 * http://opensource.org/licenses/cddl1.php for details.
 */

#include "schilyio.h"

EXPORT FILE *
#ifdef	PROTOTYPES
filemopen(const char *name, const char *mode, mode_t cmode)
#else
filemopen(name, mode, cmode)
	const char	*name;
	const char	*mode;
	mode_t		cmode;
#endif
{
	int	ret;
	int	omode = 0;
	int	flag = 0;

	if (!_cvmod(mode, &omode, &flag))
		return ((FILE *) NULL);

/*	if ((ret = _openfd(name, omode)) < 0)*/
	if ((ret = open(name, omode, cmode)) < 0)
		return ((FILE *) NULL);

	return (_fcons((FILE *)0, ret, flag));
}
