/* @(#)openfd.c	1.9 04/08/08 Copyright 1986, 1995 J. Schilling */
/*
 *	Copyright (c) 1986, 1995 J. Schilling
 */
/*
 * Copyright Jörg Schilling. All rights reserved.
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only.
 * See the file CDDL.Schily.txt in this distribution or
 * http://opensource.org/licenses/cddl1.php for details.
 */

#include "schilyio.h"

#if	defined(_openfd) && !defined(USE_LARGEFILES)
#	undef	_openfd
#endif

EXPORT int
_openfd(name, omode)
	const char	*name;
	int		omode;
{
	return (open(name, omode, 0666));
}
