/* @(#)streql.c	1.7 03/06/15 Copyright 1985, 1995-2003 J. Schilling */
/*
 *	Check if two strings are equal
 *
 *	Copyright (c) 1985, 1995-2003 J. Schilling
 */
/*
 * Copyright Jörg Schilling. All rights reserved.
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only.
 * See the file CDDL.Schily.txt in this distribution or
 * http://opensource.org/licenses/cddl1.php for details.
 */

#include <standard.h>
#include <schily.h>

EXPORT int
streql(a, b)
	const char	*a;
	const char	*b;
{
	register const char	*s1 = a;
	register const char	*s2 = b;

	if (s1 == NULL || s2 ==  NULL)
		return (FALSE);

	while (*s1 == *s2++)
		if (*s1++ == '\0')
			return (TRUE);

	return (FALSE);
}
