/* @(#)xutimes.h	1.2 03/06/12 Copyright 1996, 2003 J. Schilling */
/*
 *	Prototypes for xutimes users
 *
 *	Copyright (c) 1996, 2003 J. Schilling
 */
/*
 * Copyright Jörg Schilling. All rights reserved.
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only.
 * See the file CDDL.Schily.txt in this distribution or
 * http://opensource.org/licenses/cddl1.php for details.
 */

/*
 * star_unix.c
 */
extern	int	xutimes		__PR((char *name, struct timeval *tp));
