/* @(#)file_raise.c	1.8 04/08/08 Copyright 1986, 1995-2003 J. Schilling */
/*
 *	Copyright (c) 1986, 1995-2003 J. Schilling
 */
/*
 * Copyright Jörg Schilling. All rights reserved.
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only.
 * See the file CDDL.Schily.txt in this distribution or
 * http://opensource.org/licenses/cddl1.php for details.
 */

#include "schilyio.h"

EXPORT void
file_raise(f, flg)
	register FILE *f;
	int flg;
{
	int	oflag;
extern	int	_io_glflag;

	if (f == (FILE *)NULL) {
		if (flg)
			_io_glflag &= ~_IONORAISE;
		else
			_io_glflag |= _IONORAISE;
		return;
	}
	down(f);

	oflag = my_flag(f);

	if (flg)
		oflag &= ~_IONORAISE;
	else
		oflag |= _IONORAISE;

	set_my_flag(f, oflag);
}
